/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: OsgiMonitor.java 12246 2007-12-09 21:42:38Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.jmbeans.osgi.monitoring;

import javax.management.MBeanException;
import javax.management.Notification;
import javax.management.RuntimeOperationsException;
import javax.management.modelmbean.ModelMBean;

import org.apache.commons.modeler.BaseModelMBean;
import org.ow2.jonas.lib.jmbeans.JNotification;

/**
 * This MBean will implement the OSGI framework (services)
 * monitoring based on OSGI events.
 * @author Adriana.Danes@bull.net
 *
 */
public class OsgiMonitor {

   String objectName = null;

   // needed to send JMX notifications
   private long sequenceNumber = 0;

   private String att1 = "init";

   public OsgiMonitor(String objectName) {
       super();
       this.objectName = objectName;
   }

   /**
    * MBean interface implementation
    * @return OBJECT_NAME
    */
   public String getObjectName() {
       return objectName;
   }

   /**
    * This method will be called when the MBean detects that all the
    * required JOnAS services are correctly started.
    */
   private void notifyStartingSuccess() {
       JNotification notif = new JNotification(JNotification.startingSuccess
               , getObjectName(), sequenceNumber++);
       sendNotification(notif);
   }

   public void temporary() {
       notifyStartingSuccess();
   }

   private BaseModelMBean modelMBean = null;

   public void setModelMBean(ModelMBean modelMBean) {
       this.modelMBean = (BaseModelMBean) modelMBean;
   }

   private void sendNotification(Notification notif) {
       try {
           System.out.println("MBean " + getObjectName() + " sends notification " + notif.getType());
           modelMBean.sendNotification(notif);
       } catch (RuntimeOperationsException e) {
           // TODO Auto-generated catch block
           e.printStackTrace();
       } catch (MBeanException e) {
           // TODO Auto-generated catch block
           e.printStackTrace();
       }
   }

   public String getAtt1() {
       return att1;
   }

   public void setAtt1(String att1) {
       this.att1 = att1;
   }
}
